/*
    lavplay - Linux Audio Video PLAYback
      
    Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>


    Play MJPEG AVI-Files with the Iomega BUZ.


    Usage: lavplay <filename>



    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>


#include <linux/videodev.h>
#include "buz.h"

/* On some systems MAP_FAILED seems to be missing */

#ifndef MAP_FAILED
#define MAP_FAILED ( (caddr_t) -1 )
#endif

/* Define ZOOM_UP_QUARTER_RESOLUTION if you want to look
   at quarter resolution videos in full screen mode */

//#define ZOOM_UP_QUARTER_RESOLUTION

/* The following variables are declared in avilib.c and are
   set during avi_read_header */

extern double AVI_fps;
extern long   AVI_width;
extern long   AVI_height;
extern long   AVI_a_chans;
extern long   AVI_a_rate;
extern long   AVI_a_bits;


static int audio=0;
static int audio_rate;
static int audio_size;
static int stereo;

void audio_shutdown();

#define AUDIO_BUFFER_LEN (512*1024)
static char audio_buffer[AUDIO_BUFFER_LEN];

int get_next_frame(FILE *videofile, char *vidbuf, long vidbuflen)
{
   int res, n, alen;

   do
   {
      res = avi_read_data(videofile,vidbuf,vidbuflen,
                          audio_buffer,AUDIO_BUFFER_LEN,&alen);
      if(res==2 && audio)
      {
         n = audio_write(audio_buffer,alen,0);
         if(n<0)
         {
            fprintf(stderr,"Error writing Audio\n");
            exit(1);
         }
         if(n!=alen)
         {
            /* The audio buffer did not fit completly in the buffers
               provided by the audio system.
               In a later revision we should do something more
               sophisticated like waiting until there is space
               available - right now we just exit */
            fprintf(stderr,"Audio buffer overflow\n");
            exit(1);
         }
      }
   }
   while(res==2);

   if(res==-1) { fprintf(stderr,"Videobuffer too small\n"); exit(1); }
   if(res==-2) { fprintf(stderr,"Audiobuffer too small\n"); exit(1); }
   if(res!= 1) return EOF;

   return 0;
}

main(int argc, char ** argv)
{
   int dev;
   int res, nqueue, nsync, frame, n, hn;
   int size;
   struct buz_sync bs;
   struct buz_requestbuffers br;
   struct buz_params bp;
   char * buff;
   FILE *videofile;

   int jpg_width, jpg_height;
   unsigned char app0[16];

   if(argc != 2)
   {
      fprintf(stderr,"Usage: %s videofile\n",argv[0]);
      exit(1);
   }

   /* open video file */

   videofile = fopen(argv[1],"r");
   if(videofile==0)
   {
      perror("Open videofile");
      exit(1);
   }

   /* Read video file header */

   res = avi_read_header(videofile);
   if(res)
   {
      fprintf(stderr,"error %d reading header of %s\n",res,argv[1]);
      exit(1);
   }

   /* Set apropriate audio settings */

   switch (AVI_a_chans)
   {
      case 0: audio = 0; break;
      case 1: audio = 1; stereo = 0; break;
      case 2: audio = 1; stereo = 1; break;
      default:
         fprintf(stderr,"Number of Audio channels=%d illegal\n",AVI_a_chans);
         exit(1);
   }

   audio_rate = AVI_a_rate;
   audio_size = AVI_a_bits;


   if (audio)
   {
      printf("Audio: chans=%d rate=%d size=%d\n",AVI_a_chans,audio_rate,audio_size);
      res = audio_init(0,stereo,audio_size,audio_rate);
      if(res)
      {
         fprintf(stderr,"Error initializing Audio\n");
         exit(1);
      }
   }

   /* Ensure proper exit processing for audio */

   if(audio) atexit(audio_shutdown);

   /* open video device */

   dev = open("/dev/video", O_RDWR);
   if (dev < 0) { perror("/dev/video"); exit(1); }

   /* Request buffers */

   br.count = 32;
   br.size  = 256*1024;
   res = ioctl(dev, BUZIOC_REQBUFS,&br);
   if (res < 0) { perror("BUZIOC_REQBUFS"); exit(1); }
   printf("Got %d buffers of size %d\n",br.count,br.size);

   /* Map the buffers */
   buff = mmap(0, br.count*br.size, PROT_READ | PROT_WRITE, MAP_SHARED, dev, 0);
   if (buff == MAP_FAILED) { perror("mmap"); exit(1); }

   /* Fill all buffers first */

   for(nqueue = 0; nqueue < br.count; nqueue++)
   {
      /* Read one frame, break if EOF is reached */
      if(get_next_frame(videofile,buff+nqueue*br.size,br.size)) break;
   }

   if(nqueue == 0)
   {
      fprintf(stderr,"AVI contains no image data!\n");
      exit(1);
   }

   /* Choose the correct parameters for playback from the data in the AVI file */

   /* do a BUZIOC_G_PARAMS ioctl to get proper default values */

   res = ioctl(dev, BUZIOC_G_PARAMS, &bp);
   if (res < 0) { perror("BUZIOC_G_PARAMS"); exit(1); }

   /* Try to guess the norm (from the framerate), be generous.
      (RJ: Why for heavens sake outputs the Windows version of the Buz software
           so inaccurate framerates ?????)
      At the moment, fractions of the framerate of the norm are not supported.
    */

   fprintf(stderr,"AVI_fps = %f\n",AVI_fps);
   if (AVI_fps>23.0 && AVI_fps<27.0)
      bp.norm = 0; /* PAL */
   else if(AVI_fps>28.0 && AVI_fps<32.0)
      bp.norm = 1; /* NTSC */
   else
   {
      fprintf(stderr,"Framerate of %f not supported\n",AVI_fps);
      exit(1);
   }

   fprintf(stderr,"Norm choosen: %s\n",bp.norm?"NTSC":"PAL");

   hn = bp.norm ? 480 : 576;  /* Height of norm */

   bp.decimation = 0; /* we will set proper params ourselves */

   /* Examine the header of the first JPEG frame */

   res = check_JPEG_header (buff, br.size, &jpg_width, &jpg_height, app0);

   if(res) { fprintf(stderr,"Error in JPEG header\n"); exit(1); }

   if(jpg_width != AVI_width || (jpg_height != AVI_height && jpg_height != AVI_height/2) )
   {
      fprintf(stderr,"AVI dims: %dw x %dh, JPEG dims: %dw x %dh don't fit\n",
                     AVI_width, AVI_height, jpg_width, jpg_height);
      exit(1);
   }

   /* Check dimensions of image, select decimation factors */

   if( AVI_width > 720 || AVI_height > hn )
   {
      fprintf(stderr,"Image dimensions too big: %d x %d\n",AVI_width,AVI_height);
      exit(1);
   }
   else if(AVI_width > 360 || AVI_height > hn/2)
   {
      /* Decimation = 1 */

      if(jpg_height != AVI_height/2)
      {
         fprintf(stderr,"decimation = 1 needs 2 fields per frame\n");
         exit(1);
      }
      bp.field_per_buff = 2;
      bp.HorDcm = 1;
      bp.VerDcm = 1;
      bp.TmpDcm = 1;

      bp.img_x      = (720 - AVI_width)/2;
      bp.img_y      = (hn - AVI_height)/4;
      bp.img_width  = AVI_width;
      bp.img_height = AVI_height/2;
   }
   else
#ifdef ZOOM_UP_QUARTER_RESOLUTION
   if(AVI_width > 180 || AVI_height > hn/4)
#endif
   {
      /* Decimation = 2 */

      if(jpg_height != AVI_height)
      {
         fprintf(stderr,"decimation = 2 needs 1 field per frame\n");
         exit(1);
      }
      bp.field_per_buff = 1;
      bp.HorDcm = 2;
      bp.VerDcm = 1;
      bp.TmpDcm = 2;

      bp.img_x      = (720 - 2*AVI_width)/2;
      bp.img_y      = (hn - 2*AVI_height)/4;
      bp.img_width  = 2*AVI_width;
      bp.img_height = AVI_height;
   }
#ifdef ZOOM_UP_QUARTER_RESOLUTION
   else
   {
      /* Decimation = 4 */

      if(jpg_height != AVI_height)
      {
         fprintf(stderr,"decimation = 4 needs 1 field per frame\n");
         exit(1);
      }
      bp.field_per_buff = 1;
      bp.HorDcm = 4;
      bp.VerDcm = 2;
      bp.TmpDcm = 2;

      bp.img_x      = (720 - 4*AVI_width)/2;
      bp.img_y      = (hn - 4*AVI_height)/4;
      bp.img_width  = 4*AVI_width;
      bp.img_height = 2*AVI_height;
   }
#endif

   /* Check if the APP0 marker contains the string "AVI1".
      If so, the next byte is the "polarity" of the first field.
      If this byte is (binary) 2, we have to switch our usual
      field order.
   */

   if( strncmp(app0,"AVI1",4)==0 && app0[4]==1)
      bp.odd_even = 0;
   else
      bp.odd_even = 1;

   /* All should be set up now, set the parameters */

   res = ioctl(dev, BUZIOC_S_PARAMS, &bp);
   if (res < 0) { perror("BUZIOC_S_PARAMS"); exit(1); }


   /* Synchronize with audio (hopefully) */

   if (audio) audio_start();
   usleep(100000); /* It seems to be better with that delay */

   /* Queue the buffers read, this triggers video playback */

   for(n = 0; n < nqueue; n++)
   {
      res = ioctl(dev, BUZIOC_QBUF_PLAY, &n);
      if (res < 0) { perror("BUZIOC_QBUF_PLAY"); exit(1); }
   }

   /* Playback loop */

   nsync  = 0;

   while(1)
   {
      /* Sync to get a free buffer */

      res = ioctl(dev, BUZIOC_SYNC, &bs);
      if (res < 0) { perror("BUZIOC_SYNC"); exit(1); }
      frame = bs.frame;
      nsync++;

      /* Read one frame, break if EOF is reached */

      if(get_next_frame(videofile,buff+frame*br.size,br.size)) break;

      /* Queue the frame */

      res = ioctl(dev, BUZIOC_QBUF_PLAY, &frame);
      if (res < 0) { perror("BUZIOC_QBUF_PLAY"); exit(1); }
      nqueue++;
   }

   /* All buffers are queued, sync on the outstanding buffers */
   /* Never try to sync on the last buffer, it is a hostage of
      the codec since it is played over and over again */

   while(nqueue > nsync + 1)
   {
      res = ioctl(dev, BUZIOC_SYNC, &bs);
      if (res < 0) { perror("BUZIOC_SYNC"); exit(1); }
      nsync++;
   }

   /* Stop streaming playback */
   n = -1;
   res = ioctl(dev, BUZIOC_QBUF_PLAY, &n);
   if (res < 0) { perror("BUZIOC_QBUF_PLAY"); exit(1); }

   exit(0);
}
